from __future__ import absolute_import, division, print_function, unicode_literals
import click
import os
import re
from passthesalt import PassTheSalt, Remote, UnauthorizedAccess, ConflictingTimestamps, \
                        to_clipboard, generate, __version__

click.disable_unicode_literals_warning = True

PATH = os.path.expanduser('~/.passthesalt')
REMOTE = os.path.expanduser('~/.passthesalt.remote')


class URLParamType(click.ParamType):
    name = 'url'

    def convert(self, value, param, ctx):
        temp = re.sub(r'https?://', '', value).rstrip('/')
        if '.' in temp[1:-1]:
            return temp
        else:
            self.fail('"{}" is not a valid url'.format(value), param, ctx)


URL = URLParamType()


def get_master_password(pts):
    master_password = click.prompt('Enter master password', hide_input=True)
    for i in range(2):
        if pts.master_valid(master_password):
            break
        click.echo('Error: invalid master password.')
        master_password = click.prompt('Enter master password', hide_input=True)
    else:
        click.echo('Error: three incorrect attempts.')
        raise click.Abort()
    return master_password


def renew_remote_token(remote):
    click.echo('Unauthorized access. Need to renew token:')
    name = click.prompt('Enter name')
    password = click.prompt('Enter password', hide_input=True)
    remote.renew_token(name, password)
    remote.save(REMOTE)


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(__version__, '-v', '--version', prog_name='passthesalt', message='%(prog)s %(version)s')
@click.pass_context
def cli(ctx):
    """
    \b
     ____                 _   _            ____        _ _
    |  _ \ __ _ ___ ___  | |_| |__   ___  / ___|  __ _| | |_
    | |_) / _` / __/ __| | __| '_ \ / _ \ \___ \ / _` | | __|
    |  __/ (_| \__ \__ \ | |_| | | |  __/  ___) | (_| | | |_
    |_|   \__,_|___/___/  \__|_| |_|\___| |____/ \__,_|_|\__|

    A deterministic password generation and password storage system.
    """
    pts = PassTheSalt()
    remote = Remote()

    if os.path.isfile(PATH):
        pts.load(PATH)
    else:
        click.echo('Initializing Pass the Salt ...')
        owner = click.prompt('Please enter your name')

        if click.confirm('Use master password validation?'):
            master_password = click.prompt('Please enter the master password',
                                           confirmation_prompt=True, hide_input=True)
            pts.initialize(owner, master_password)
        else:
            pts.initialize(owner)

        pts.save(PATH)
        click.echo('Successfully initialized Pass The Salt!\n')

    if os.path.isfile(REMOTE):
        remote = remote.load(REMOTE)
    elif ctx.invoked_subcommand in ('push', 'pull'):
        click.echo('Initializing remote ...')

        url = click.prompt('Enter GET and PUT url')
        token_url = click.prompt('Enter token renewal url')
        remote.initialize(url, token_url)

        remote.save(REMOTE)
        click.echo('Successfully initialized remote config.')

    ctx.obj = {
        'pts': pts,
        'remote': remote
    }


@cli.command('add')
@click.argument('label', required=False)
@click.option('--encrypt', '-e', is_flag=True, help='Store and encrypt a secret.')
@click.option('--clipboard/--no-clipboard', default=True,
              help='Whether to copy secret to clipboard or print it out (default: clipboard).')
@click.pass_obj
def pts_add(obj, label, encrypt, clipboard):
    """
    Store secret.

    LABEL is the unique label for the secret.
    """
    pts = obj['pts']

    if not label:
        label = click.prompt('Enter label for new secret')

    if pts.exists(label):
        click.echo('Label "{}" already exists.'.format(label))
        click.confirm('Update?', abort=True)

    master_password = None
    if encrypt:
        secret = click.prompt('Enter secret to store', confirmation_prompt=True, hide_input=True)
        master_password = get_master_password(pts)
        pts.store_encrypted(label, secret, master_password)
    else:
        domain = click.prompt('Enter domain name', type=URL)
        username = click.prompt('Enter username')
        iteration = click.prompt('Enter iteration', default='0')
        salt = '|'.join([domain, username, iteration])
        if click.confirm('Store "{}" as "{}"'.format(salt, label), abort=True):
            pts.store_generatable(label, salt)
    pts.save(PATH)
    click.echo('Secret stored!')

    if not encrypt and click.confirm('\nRetrieve secret?'):
        if not master_password:
            master_password = get_master_password(pts)

        secret = pts.get(label, master_password)
        if clipboard:
            to_clipboard(secret, timeout=10)
            click.echo('Secret copied to clipboard.')
        else:
            click.echo(secret)


@cli.command('get')
@click.argument('label', required=False)
@click.option('--salt', '-s',
              help='Run the secret generation algorithm on the given description.')
@click.option('--clipboard/--no-clipboard', default=True,
              help='Whether to copy password to clipboard or print it out (default: clipboard).')
@click.pass_obj
def pts_get(obj, label, salt, clipboard):
    """
    Retrieve secret.

    LABEL is the unique label of the secret.
    """
    pts = obj['pts']

    if salt:
        master_key = pts.master_key(get_master_password(pts))
        secret = generate(salt, master_key)
    else:
        if not label:
            label = click.prompt('Enter label of secret to retrieve')

        if pts.exists(label):
            master_password = get_master_password(pts)
            secret = pts.get(label, master_password)
        else:
            click.echo('Label "{}" does not exist.'.format(label))
            raise click.Abort()

    if clipboard:
        to_clipboard(secret, timeout=10)
        click.echo('Secret copied to clipboard.')
    else:
        click.echo(secret)


@cli.command('ls')
@click.argument('label', required=False)
@click.option('--type', '-t', type=click.Choice(['encrypted', 'generatable']))
@click.option('--verbose', '-v', count=True, help='More information.')
@click.option('--quiet', '-q', is_flag=True, help='Do not print anything if no secrets.')
@click.pass_obj
def pts_list(obj, label, type, verbose, quiet):
    """
    List the secrets.
    """
    pts = obj['pts']

    subset = [l for l in pts.labels if not label or l.lower().startswith(label.lower())]

    if len(subset) == 0:
        if not quiet:
            click.echo('No stored secrets.')
        return

    col = max(len(l) for l in subset) + 4

    for label in sorted(subset):
        if not type or type == pts.labels[label]['type']:
            if verbose > 1 and label in pts.generatable:
                click.echo('{:<{col}}{:<{col_}}{}'.format(
                    label,
                    'type: ' + pts.labels[label]['type'],
                    'salt: ' + pts.generatable[label],
                    col=col,
                    col_=20
                ))
            elif verbose > 0:
                click.echo('{:<{col}}{}'.format(
                    label,
                    'type: ' + pts.labels[label]['type'],
                    col=col
                ))
            else:
                click.echo(label)


@cli.command('rm')
@click.argument('label', required=False)
@click.pass_obj
def pts_remove(obj, label):
    """
    Remove secret.

    LABEL is the unique label of the secret.
    """
    pts = obj['pts']

    if not label:
        label = click.prompt('Enter label of secret to remove')

    if not pts.exists(label):
        click.echo('Label "{}" does not exist.'.format(label))
        raise click.Abort()

    click.confirm('Are you sure you want to remove "{}"?'.format(label), abort=True)

    info = pts.labels[label]

    if info['type'] == 'generatable':
        pts.remove_generatable(label)
    elif info['type'] == 'encrypted':
        master_password = get_master_password(pts)
        pts.remove_encrypted(label, master_password)

    pts.save(PATH)
    click.echo('Successfully removed "{}".'.format(label))


@cli.command('mv')
@click.argument('label')
@click.argument('new-label')
@click.pass_obj
def pts_move(obj, label, new_label):
    """
    Rename secret.

    LABEL is the unique name of the secret.
    NEW-LABEL is a new unique name for the secret..
    """
    pts = obj['pts']

    if not pts.exists(label):
        click.echo('Label "{}" does not exist.'.format(label))
        raise click.Abort()

    if pts.exists(new_label):
        click.echo('New label already exists.'.format(new_label))
        raise click.Abort()

    info = pts.labels[label]

    if info['type'] == 'generatable':
        pts.rename_generatable(label, new_label)
    elif info['type'] == 'encrypted':
        master_password = get_master_password(pts)
        pts.rename_encrypted(label, new_label, master_password)

    pts.save(PATH)
    click.echo('Successfully renamed "{}" to "{}".'.format(label, new_label))


@cli.command('push')
@click.option('--force', '-f', is_flag=True, help='Ignore timestamp conflicts.')
@click.pass_obj
def pts_push(obj, force):
    """
    Update remote store with local store.
    """
    pts = obj['pts']
    remote = obj['remote']

    for _ in range(2):
        try:
            remote.put(pts, force=force)
            break
        except UnauthorizedAccess:
            renew_remote_token(remote)
        except ConflictingTimestamps:
            click.echo('Timestamp conflict. The server has a newer version of the store. Use --force to override.')
            raise click.Abort()

    remote.touch()
    remote.save(REMOTE)
    click.echo('Successfully pushed to remote store.')


@cli.command('pull')
@click.option('--force', '-f', is_flag=True, help='Ignore timestamp conflicts.')
@click.pass_obj
def pts_pull(obj, force):
    """
    Update local store with remote store.
    """
    pts = obj['pts']
    remote = obj['remote']

    for _ in range(2):
        try:
            remote_pts, modified = remote.get()
            break
        except UnauthorizedAccess:
            renew_remote_token(remote)

    if force or (hasattr(pts, 'modified') and modified > pts.modified):
        remote_pts.save(PATH)
        click.echo('Successfully pulled remote store.')
    else:
        click.echo('Timestamp conflict. The local version is newer than the remote.')
        raise click.Abort()


if __name__ == '__main__':
    cli()



from __future__ import absolute_import, division, print_function, unicode_literals
from binascii import unhexlify, hexlify as hexlify_
import Crypto.Cipher.AES
import base64
import datetime
import dateutil.parser
import hashlib
import json
import os
import pyperclip
import requests
import subprocess
import sys


__version__ = '2.1.0'


if sys.version_info < (3, 0):
    def shiftlify(bytes, chars, shift=0):
        bytes = [(ord(byte) + shift) % 256 for byte in bytes]
        return ''.join(chars[byte % len(chars)] for byte in bytes)
else:
    def shiftlify(bytes, chars, shift=0):
        bytes = [(byte + shift) % 256 for byte in bytes]
        return ''.join(chars[byte % len(chars)] for byte in bytes)


def passlify(bytes):
    """
    Create a password from a byte list.
    """
    lower = 'abcdefghijklmnopqrstuvwxyz'
    upper = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    numbers = '0123456789'
    symbols = '!@#$%^&*_+-='
    chars = lower + upper + numbers + symbols

    result = shiftlify(bytes, chars)

    # Attempt to make sure the password contains:
    #   a lowercase letter
    #   an uppercase letter
    #   a number
    #   a symbol (one of !@#$%^&*_+-=)
    for _ in range(256):
        if any(c in lower for c in result) and any(c in upper for c in result) and \
           any(c in numbers for c in result) and any(c in symbols for c in result):
            break
        result = shiftlify(bytes, chars, shift=1)

    # Attempt to make sure the password starts with a letter
    for _ in range(len(result)):
        if result[0] in lower + upper:
            break
        result = result[1:] + result[0:1]

    return result


def hexlify(bytes):
    return hexlify_(bytes).decode('utf-8')


def encrypt(dictionary, master_key):
    """
    AES (CFB) encrypt 'dictionary' with 'master_key'.
    """
    key = hashlib.sha1(master_key.encode()).hexdigest()[:32]
    iv = os.urandom(16)
    aes_obj = Crypto.Cipher.AES.new(key, Crypto.Cipher.AES.MODE_CFB, iv)
    return hexlify(iv + aes_obj.encrypt(json.dumps(dictionary, sort_keys=True)))


def decrypt(encoded, master_key):
    """
    Decrypt an AES (CFB) encrypted dictionary.
    """
    key = hashlib.sha1(master_key.encode()).hexdigest()[:32]
    iv = unhexlify(encoded[:32])
    aes_obj = Crypto.Cipher.AES.new(key, Crypto.Cipher.AES.MODE_CFB, iv)
    return json.loads(aes_obj.decrypt(unhexlify(encoded[32:])))


def pbkdf2_hash(password, salt=None):
    """
    PBKDF2 HMAC-SHA-256 hash a password (100 000 iterations)
    """
    if not salt:
        salt = os.urandom(16)
    else:
        salt = unhexlify(salt)

    hash_ = hashlib.pbkdf2_hmac(
        'sha256',
        password.encode(),
        salt,
        100000,
        dklen=20
    )

    return hexlify(salt), hexlify(hash_)


def generate(salt, master_key):
    """
    Generate a password from a master key and password description.
    """
    return passlify(hashlib.pbkdf2_hmac(
        'sha256',
        master_key.encode(),
        salt.encode(),
        2048,
        dklen=20
    ))


def to_clipboard(text, timeout=None):
    """
    Copy text to clipboard.
    """
    pyperclip.copy(text)
    if timeout:
        command = 'sleep {} && {} -c "import pyperclip;pyperclip.copy(\'\');"'.format(timeout, sys.executable)
        subprocess.Popen(command, stdin=subprocess.PIPE, shell=True)


class RemoteError(Exception):
    pass


class ConflictingTimestamps(RemoteError):
    pass


class UnauthorizedAccess(RemoteError):
    pass


class ServerError(RemoteError):
    pass


class File(object):

    def to_dict(self):
        d = {k: v for k, v in vars(self).items() if v}
        if 'modified' in d:
            d['modified'] = d['modified'].isoformat()
        return d

    def load(self, path):
        with open(path, 'r') as f:
            self.loads(json.load(f))
        return self

    def loads(self, data):
        for key in data:
            if key == 'modified':
                value = dateutil.parser.parse(data[key])
            else:
                value = data[key]
            setattr(self, key, value)
        return self

    def save(self, path):
        with open(path, 'w') as f:
            json.dump(self.to_dict(), f, sort_keys=True, indent=2)

    def dumps(self, indent=None):
        return json.dumps(self.to_dict(), sort_keys=True, indent=indent)

    def encode(self):
        return base64.b64encode(self.dumps().encode()).decode('ascii')

    def decode(self, data):
        return self.loads(json.loads(base64.b64decode(data.encode()).decode('utf-8')))

    def touch(self):
        self.modified = datetime.datetime.utcnow()


class Remote(File):

    def __init__(self):
        self.remote = dict()

    def initialize(self, url, token_url):
        self.remote['url'] = url
        self.remote['token_url'] = token_url

    def request(self, verb, url=None, auth=None, data=None):
        headers = {'Content-Type': 'application/json'}
        if auth is None:
            auth = requests.auth.HTTPBasicAuth(self.remote.get('token'), 'unused')
        if url is None:
            url = self.remote['url']

        response = requests.request(verb, url, headers=headers, auth=auth, data=data)

        if response.status_code == 401:
            raise UnauthorizedAccess(response.json().get('message'))
        elif response.status_code == 409:
            raise ConflictingTimestamps(response.json().get('message'))
        elif 500 > response.status_code >= 300:
            raise RemoteError(response.json())
        elif response.status_code >= 500:
            raise ServerError(response.text)

        return response.json()

    def get(self):
        data = self.request('GET')
        pts = PassTheSalt().decode(data['value'])
        modified = dateutil.parser.parse(data['modified'])
        return pts, modified

    def put(self, pts, force=False):
        payload = {'value': pts.encode()}
        if not force:
            modifieds = list()
            if hasattr(self, 'modified'):
                modifieds.append(self.modified)
            if hasattr(pts, 'modified'):
                modifieds.append(pts.modified)
            if modifieds:
                payload['modified'] = max(modifieds).isoformat()
        data = json.dumps(payload)
        self.request('PUT', data=data)

    def renew_token(self, name, password):
        auth = requests.auth.HTTPBasicAuth(name, password)
        data = self.request('GET', self.remote['token_url'], auth=auth)
        self.remote['token'] = data['token']


class PassTheSalt(File):

    def __init__(self):
        self.config = dict()
        self.labels = dict()
        self.generatable = dict()
        self.encrypted = None

    def initialize(self, owner, master_password=None):
        self.config['owner'] = owner
        if master_password:
            salt, hash_ = pbkdf2_hash(master_password)
            self.config['master'] = {
                'hash': hash_,
                'salt': salt
            }

    def master_key(self, master_password):
        return self.config['owner'] + '|' + master_password

    def master_valid(self, master_password):
        if 'master' in self.config:
            _, hash_ = pbkdf2_hash(master_password, self.config['master']['salt'])
            return hash_ == self.config['master']['hash']
        else:
            return True

    def exists(self, label):
        return label in self.labels

    def get(self, label, master_password):
        master_key = self.master_key(master_password)
        info = self.labels[label]
        if info['type'] == 'generatable':
            return generate(self.generatable[label], master_key)
        elif info['type'] == 'encrypted':
            return decrypt(self.encrypted, master_key)[label]

    def _store(self, label, type_):
        self.touch()
        self.labels[label] = {
            'type': type_,
            'modified': datetime.date.today().strftime("%Y%m%d")
        }

    def _remove(self, label):
        self.touch()
        del self.labels[label]

    def _rename(self, label, new_label):
        self.touch()
        self.labels[new_label] = self.labels.pop(label)

    def store_generatable(self, label, salt):
        self._store(label, 'generatable')
        self.generatable[label] = salt

    def remove_generatable(self, label):
        self._remove(label)
        del self.generatable[label]

    def rename_generatable(self, label, new_label):
        self._rename(label, new_label)
        self.generatable[new_label] = self.generatable[label]

    def store_encrypted(self, label, secret, master_password):
        master_key = self.master_key(master_password)
        if self.encrypted is not None:
            decrypted = decrypt(self.encrypted, master_key)
        else:
            decrypted = dict()
        self._store(label, 'encrypted')
        decrypted[label] = secret
        self.encrypted = encrypt(decrypted, master_key)

    def remove_encrypted(self, label, master_password):
        master_key = self.master_key(master_password)
        decrypted = decrypt(self.encrypted, master_key)
        self._remove(label)
        del decrypted[label]
        if not decrypted:
            self.encrypted = None
        else:
            self.encrypted = encrypt(decrypted, master_key)

    def rename_encrypted(self, label, new_label, master_password):
        master_key = self.master_key(master_password)
        decrypted = decrypt(self.encrypted, master_key)
        self._rename(label, new_label)
        decrypted[new_label] = decrypted.pop(label)
        self.encrypted = encrypt(decrypted, master_key)



import io
import os
import re
from setuptools import setup


def read(*path):
    """
    Python 2/3 file reading.
    """
    version_file = os.path.join(os.path.dirname(__file__), *path)
    with io.open(version_file, encoding='utf8') as f:
        return f.read()


def find_version():
    """
    Regex search __init__.py so that we do not have to import.
    """
    text = read('passthesalt', '__init__.py')
    match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', text, re.M)
    if match:
        return match.group(1)
    raise RuntimeError('Unable to find version string.')


version = find_version()

long_description = read('README.rst')

install_requires = [
    'click==6.7',
    'pycrypto==2.6.1',
    'pyperclip==1.5.27',
    'python-dateutil==2.6.1',
    'requests==2.18.4'
]

entry_points = {
    'console_scripts': [
        'pts=passthesalt.__main__:cli',
        'passthesalt=passthesalt.__main__:cli'
    ]
}

setup(
    name='passthesalt',
    packages=['passthesalt'],
    version=version,
    install_requires=install_requires,
    entry_points=entry_points,
    python_requires='>=2.7',
    description='Deterministic password generation and password storage.',
    long_description=long_description,
    author='Ross MacArthur',
    author_email='macarthur.ross@gmail.com',
    license='MIT',
    url='https://github.com/rossmacarthur/passthesalt',
    download_url='https://github.com/rossmacarthur/passthesalt/archive/{}.tar.gz'.format(version),
    keywords='password manager pbkdf2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
)
